/*
Copyright 2023.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package controller

import (
	"context"

	apierrors "k8s.io/apimachinery/pkg/api/errors"
	"k8s.io/apimachinery/pkg/api/resource"
	"k8s.io/apimachinery/pkg/runtime"
	ctrl "sigs.k8s.io/controller-runtime"
	"sigs.k8s.io/controller-runtime/pkg/client"
	"sigs.k8s.io/controller-runtime/pkg/log"

	databasev1 "github.com/iwanhae/nodb/api/v1"
	"github.com/iwanhae/nodb/internal/templates"
	"github.com/pkg/errors"
	corev1 "k8s.io/api/core/v1"
)

// PostgreSQLReconciler reconciles a PostgreSQL object
type PostgreSQLReconciler struct {
	client.Client
	Scheme *runtime.Scheme
}

//+kubebuilder:rbac:groups=database.iwanhae.kr,resources=postgresqls,verbs=get;list;watch;create;update;patch;delete
//+kubebuilder:rbac:groups=database.iwanhae.kr,resources=postgresqls/status,verbs=get;update;patch
//+kubebuilder:rbac:groups=database.iwanhae.kr,resources=postgresqls/finalizers,verbs=update
//+kubebuilder:rbac:groups="",resources=pods,verbs=get;create

// Reconcile is part of the main kubernetes reconciliation loop which aims to
// move the current state of the cluster closer to the desired state.
// TODO(user): Modify the Reconcile function to compare the state specified by
// the PostgreSQL object against the actual cluster state, and then
// perform operations to make the cluster state reflect the state specified by
// the user.
//
// For more details, check Reconcile and its Result here:
// - https://pkg.go.dev/sigs.k8s.io/controller-runtime@v0.16.0/pkg/reconcile
func (r *PostgreSQLReconciler) Reconcile(ctx context.Context, req ctrl.Request) (ctrl.Result, error) {
	logger := log.FromContext(ctx)

	obj := databasev1.PostgreSQL{}
	if err := r.Get(ctx, req.NamespacedName, &obj); err != nil {
		if client.IgnoreNotFound(err) == nil {
			return ctrl.Result{}, nil
		}
		logger.Error(err, "resource not found")
		return ctrl.Result{}, err
	}

	logger.Info("reconcile", "namespace", obj.Namespace, "name", obj.Name)

	// Pending: Creating Pod if not exists
	// do not handle spec update

	pod := corev1.Pod{}
	if err := r.Client.Get(ctx, req.NamespacedName, &pod); err == nil {
		// if found return
		logger.Info("ignore already existing pod")
		return ctrl.Result{}, nil
	} else if !apierrors.IsNotFound(err) {
		// weird error
		return ctrl.Result{}, err
	}

	pod = templates.PostgreSQLPod(templates.PostgreSQLOpts{
		Name:      obj.Name,
		Namespace: obj.Namespace,
		Tag:       obj.Spec.Version,

		User:     obj.Spec.User,
		Password: obj.Spec.Password,
		Database: obj.Spec.Database,

		Memory: resource.MustParse("1Gi"),
		Owner:  &obj,
	})

	logger.Info("create pod", "namespace", pod.Namespace, "name", pod.Name)
	if err := r.Client.Create(ctx, &pod); err != nil {
		return ctrl.Result{}, errors.Wrap(err, "failed to create pod")
	}

	obj.Status.Status = databasev1.Status_Initializing
	logger.Info("update status")
	if err := r.Status().Update(ctx, &obj); err != nil {
		return ctrl.Result{}, errors.Wrap(err, "failed to patch status")
	}

	return ctrl.Result{}, nil
}

// SetupWithManager sets up the controller with the Manager.
func (r *PostgreSQLReconciler) SetupWithManager(mgr ctrl.Manager) error {
	return ctrl.NewControllerManagedBy(mgr).
		For(&databasev1.PostgreSQL{}).
		Complete(r)
}
